"""Module containing the rumdSimulation class"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


import time
import weakref
# the following requires python2.7
from collections import OrderedDict
# if python 2.7 is not installed, install an equivalent implementation from
# https://pypi.python.org/pypi/ordereddict
# Then comment out the above import and uncomment
#from ordereddict import OrderedDict

from . import rumdswig as rumd
from .pythonOutputManager import pythonOutputManager

class Simulation(object):

    """
    Contains the main objects and provides basic simulation main loop.
    Once created at minimum the potential and integrator must be set
    before Run() can be called.

    """

    def __init__(self, filename, pb=0, tp=0, verbose=True):
        """
        Create the simulation object, and the underlying sample object, after
        reading in a configuration from the named file. Assign default values
        to various simulation- and output-related parameters. Parameters pb
        and tp set the number of particles per block and threads per particle
        respectively; if not present, default values (depending on system size)
        will be chosen. verbose=False suppresses messages to standard output.

        """
        self.sample = rumd.Sample(pb, tp)
        self.sample.SetVerbose(verbose)
        self.sample.ReadConf(filename)

        self.simulationBox = None
        self.thermostat = False
        self.itg = None
        self.potentialList = []
        self.output_is_initialized = False
        self.blockSize = None
        self.moleculeData = None
        self.verbose = verbose

        self.outputParameters = {}
        self.outputParameters["energies"] = ("linear", 256)
        self.outputParameters["trajectory"] = ("logarithmic", 1)
        self.allowedManagers = ["energies", "trajectory"]
        self.pyOutputList = []
        self.outputDirectory = self.sample.GetOutputDirectory()

        self.runtime_actions = {}
        # use ordered dict so we can control the order of iteration
        self.runtime_action_intervals = OrderedDict()
        # need weak references to these functions to avoid reference loops.
        # Cannot make a weak reference to a bound method, so need to store
        # separate weak references to the instance object and function object
        self.runtime_actions["momentum_reset"] = (
            weakref.ref(self),
            weakref.ref(self.ResetMomentum.__func__))
        self.runtime_action_intervals["momentum_reset"] = 100

        self.runtime_actions["sort"] = (
            weakref.ref(self.sample),
            weakref.ref(self.sample.SortParticles.__func__))
        self.runtime_action_intervals["sort"] = 0



        self.nominalNumBlocks = 100
        self.exclusionInfo = {"Bond":set(),
                              "Angle":False,
                              "Dihedral":False}

        self.timing_filename = "rumd_timing_info.dat"
        device_info = rumd.Device.GetDevice().GetDeviceReport()
        with open(self.timing_filename, "w") as timing_file:
            timing_file.write("# %s \n" % device_info)
            timing_file.write("# nsteps, time(s), MATS\n")

        self.last_time_stamp = time.time() # rumd.Device.GetDevice().Time() #
        self.elapsed = 0.
        self.write_timing_info = True

    def SetVerbose(self, vb):
        """
        Turn on/off most messages in rumdSimulation and sample objects.
        sim.SetVerbose automatically calls sim.sample.SetVerbose; call
        to the latter explicitly to only affect sample messages.
        """
        self.verbose = vb
        self.sample.SetVerbose(self.verbose)

    def SetPB_TP(self, pb, tp):
        """
        Change pb and tp for the sample. Used mainly by the autotuner.
        Involves copying of data due array re-allocation.
        """
        self.sample.SetPB_TP(pb, tp)
        if self.moleculeData is not None:
            self.RecreateExclusionList()
        self.sample.UpdateDegreesOfFreedom()

    def WritePotentials(self):
        """
        Write a file containing the pair potential
        and pair force for each associated pair potential.
        """
        for potential in self.potentialList:
            potential.WritePotentials(self.sample.GetSimulationBox())


    def WriteConstraints(self):
        """
        Write a file listing the deviations of bond lengths from
        constrained values.
        """
        if self.moleculeData is None:
            raise ValueError("No MoleculeData object present")
        self.moleculeData.WriteConstraints()


    def ReadMoleculeData(self, topologyFile):
        """
        Read data describing how particles are connected as molecules
        and create the MoleculeData object.
        """
        self.moleculeData = rumd.MoleculeData(self.sample, topologyFile)
        self.sample.SetMoleculeData(self.moleculeData)

    def SetExclusionBond(self, bond_type):
        """
        Exclude bonds of a given type. Not necessary to be called by user.
        """
        if self.verbose:
            print("Setting exclusion bond, bond_type=", bond_type)
        for potential in self.potentialList:
            self.moleculeData.SetExclusionBond(potential, bond_type)

    def SetBondHarmonic(self, bond_type, lbond, ks, exclude=True):
        """
        Set a bond-type to be harmonic.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData " \
                  " first")
        self.moleculeData.SetBondHarmonic(bond_type, lbond, ks)
        if exclude:
            self.exclusionInfo["Bond"].add(bond_type)
            self.SetExclusionBond(bond_type)
        else:
            self.exclusionInfo["Bond"].discard(bond_type)


    def SetBondFENE(self, bond_type, max_l, k):
        """
        Set a bond-type to be the FENE type.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData " \
                  "first")
        self.moleculeData.SetBondFENE(bond_type, max_l, k)

    def SetExclusionAngle(self):
        """
        Exclude pair interactions associated with bond-angle interactions.
        Not necessary to be called by user.
        """
        for potential in self.potentialList:
            self.moleculeData.SetExclusionAngle(potential)

    def SetAngleCosSq(self, angle_type, theta0, ktheta, exclude=True):
        """
        Set angle type coefficients.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData" \
                  " first")
        self.moleculeData.SetAngleCosSq(angle_type, theta0, ktheta)

        if exclude:
            self.exclusionInfo["Angle"] = True
            self.SetExclusionAngle()
        else:
            self.exclusionInfo["Angle"] = False

    def SetAngleSq(self, angle_type, theta0, ktheta, exclude=True):
        """
        Set angle type coefficients.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData first")
        self.moleculeData.SetAngleSq(angle_type, theta0, ktheta)

        if exclude:
            self.exclusionInfo["Angle"] = True
            self.SetExclusionAngle()
        else:
            self.exclusionInfo["Angle"] = False



    def GetAngles(self):
        """
        Access array containing current values of bond angles.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData" \
                  " first")
        self.moleculeData.GetAngles()

    def GetDihedrals(self):
        """
        Access array containing current values of dihedral angles.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData" \
                  " first")
        self.moleculeData.GetDihedrals()

    def SetExclusionDihedral(self):
        """
        Exclude pair interactions associated with dihedral angle interactions.
        Not necessary to be called by user.
        """


        for potential in self.potentialList:
            self.moleculeData.SetExclusionDihedral(potential)

    def SetDihedralRyckaert(self, dihedral_type, coeffs, exclude=True):
        """
        Set coefficients for a given type of dihedral angle, by default
        excluding corresponding pair interactions.
        """
        if len(coeffs) != 6:
            raise ValueError("The coefficient array must have length 6")

        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData" \
                  " first")
        self.moleculeData.SetDihedralRyckaert(dihedral_type,
                                              coeffs[0], coeffs[1],
                                              coeffs[2], coeffs[3],
                                              coeffs[4], coeffs[5])
        if exclude:
            self.exclusionInfo["Dihedral"] = True
            self.SetExclusionDihedral()
        else:
            self.exclusionInfo["Dihedral"] = False

    def SetPeriodicDihedral(self, dihedral_type, coeffs, exclude=True):
        """
        Set coefficients for a given type of dihedral angle, by default
        excluding corresponding pair interactions.
        """
        if len(coeffs) != 4:
            raise ValueError("The coefficient array must have length 4")
        print("Periodic")

        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData first")
        self.moleculeData.SetPeriodicDihedral(dihedral_type, coeffs[0],
                                              coeffs[1], coeffs[2], coeffs[3])
        if exclude:
            self.exclusionInfo["Dihedral"] = True
            self.SetExclusionDihedral()
        else:
            self.exclusionInfo["Dihedral"] = False


    def SetBondConstraint(self, bond_type, lbond, exclude=True):
        """
        Set a bond-type as "constraint" and pass the length, by default
        excluding interactions from the pair potential for bonds of this type.
        """
        if self.moleculeData is None:
            raise ValueError("No moleculeData object; call ReadMoleculeData" \
                  " first")

        self.moleculeData.SetBondConstraint(bond_type, lbond)
        if exclude:
            self.exclusionInfo["Bond"].add(bond_type)
            self.SetExclusionBond(bond_type)
        else:
            self.exclusionInfo["Bond"].discard(bond_type)

    def RecreateExclusionList(self):
        """
        Recreates the exclusion list via stored exclusion information. Not
        necessary to be called by user.
        """
        for bond_type in self.exclusionInfo["Bond"]:
            self.SetExclusionBond(bond_type)
        if self.exclusionInfo["Angle"]:
            self.SetExclusionAngle()
        if self.exclusionInfo["Dihedral"]:
            self.SetExclusionDihedral()

    def ScaleSystem(self, scaleFactor, direction=None, CM=False):
        """
        Scale the simulation box and particle positions by scaleFactor
        in all directions, if direction is ommitted; otherwise only
        in the direction specified by direction.
        For molecular systems, the centers of mass can be scaled
        while keeping intra molecular distances constant.
        """
        if direction is None and not CM:
            self.sample.IsotropicScaleSystem(scaleFactor)
        elif direction is None and CM:
            self.sample.IsotropicScaleSystemCM(scaleFactor)
        elif not CM:
            self.sample.AnisotropicScaleSystem(scaleFactor, direction)
        else:
            raise ValueError("Anisotropic scaling of the molecular center" \
                  " of mass is not implemented yet")

    def SetSimulationBox(self, simBox):
        """
        Set an alternative simulationBox.
        """
        self.simulationBox = simBox # keep it alive by having a reference
        self.sample.SetSimulationBox(simBox)

    def SetPotential(self, potential):
        """
        Set a potential for this system, replacing any previous
        potentials.
        """
        if len(self.potentialList) > 0:
            print("[Info] rumdSimulation.SetPotential(): previously assigned" \
                  " potentials will be replaced. Use AddPotential to keep" \
                  " more than one potential.")
        self.potentialList = [potential]
        self.sample.SetPotential(potential)

    def AddPotential(self, potential):
        """
        Add a potential for this system, keeping any previous potentials
        (the total force vector and potential will be the sum over all
        added potentials).
        """
        self.potentialList.append(potential)
        self.sample.AddPotential(potential)

        if self.moleculeData is not None:
            for bond_type in self.exclusionInfo["Bond"]:
                self.moleculeData.SetExclusionBond(potential, bond_type)
            if self.exclusionInfo["Angle"]:
                self.moleculeData.SetExclusionAngle(potential)
            if self.exclusionInfo["Dihedral"]:
                self.moleculeData.SetExclusionDihedral(potential)



    def SetIntegrator(self, itg):
        """
        Set the integrator for this system, and call its momentum-resetting
        function.
        """
        self.sample.SetIntegrator(itg)
        self.itg = itg

    def WriteConf(self, filename):
        """
        Write the current configuration to a file.
        """
        self.sample.WriteConf(filename)

    def SetBlockSize(self, blockSize):
        """
        Specify the block-size for output (None means automatic). Should be a
        power of 2 for logarithmic scheduling.
        """
        self.blockSize = blockSize

    def SetOutputDirectory(self, outputDirectory):
        """
        Change the directory for output files, default TrajectoryFiles.
        """
        self.outputDirectory = outputDirectory
        self.sample.SetOutputDirectory(outputDirectory)

    def GetNumberOfParticles(self):
        """
        Return the number of particles in the sample object.
        """
        return self.sample.GetNumberOfParticles()

    def GetVolume(self):
        """
        Return the volume defined by the simulation box.
        """
        return self.sample.GetSimulationBox().GetVolume()


    def SetOutputMetaData(self, manager_name, **kwargs):
        """
        Access to the output manager to control precision and what gets written.

        Examples
        ---------

        sim.SetOutputMetaData("trajectory", precision=6, virials=True)
        sim.SetOutputMetaData("energies", potentialEnergy=False)
        """
        if manager_name not in self.allowedManagers:
            raise ValueError("SetOutputMetaData -- first argument must be " \
                  "one of" + str(self.allowedManagers))

        if manager_name in self.pyOutputList:
            raise ValueError("SetOutputMetaData not implemented for python" \
                  " output managers")

        for key in kwargs:
            self.sample.SetOutputManagerMetaData(manager_name, key, kwargs[key])

    def AddExternalCalculator(self, calc):
        """
        Add data computed by an external calculator class to the energies files.

        Example
        -------

        alt_pot_calc = AlternatePotentialCalculator(...)
        sim.AddExternalCalculator(alt_pot_calc)

        """
        self.sample.AddExternalCalculator(calc)

    def RemoveExternalCalculator(self, calc):
        """
        Remove/disassociate the calculator object calc from the energies output
        manager
        """
        self.sample.RemoveExternalCalculator(calc)


    def AddOutputManager(self, manager_name, manager_obj):
        """
        Add an existing output manager object (typically a C++ output
        manager), specifying a name which will be used to refer to the
        object when calling SetOutputScheduling and SetOutputMetaData.
        The name can be anything, but the convention is that it matches
        the names of the output files associated with this manager.
        """
        if manager_name in self.allowedManagers:
            raise ValueError("Output manager name %s is " \
                  "already in use" % manager_name)
        self.allowedManagers.append(manager_name)
        self.outputParameters[manager_name] = ("logarithmic", 1)
        self.sample.AddOutputManager(manager_name, manager_obj)


    def NewOutputManager(self, manager_name):
        """
        Create a new (python) output manager, which will write files in the
        output directory starting with manager_name.
        """
        if manager_name in self.allowedManagers:
            raise ValueError("Output manager name %s is" \
                  " already in use" % manager_name)
        self.pyOutputList.append(
            pythonOutputManager(manager_name, self.outputDirectory))
        self.allowedManagers.append(manager_name)

    def SetOutputParameters(self, manager_name, schedule, **kwargs):
        """
        Old (deprecated) name for SetOutputScheduling.
        """
        print("[Info]: SetOutputParameters is deprecated. Use " \
              "SetOutputScheduling instead")
        self.SetOutputScheduling(manager_name, schedule, **kwargs)


    def SetOutputScheduling(self, manager_name, schedule, **kwargs):
        """
        Set scheduling information for an output manager.

        manager must be one of the current managers, which include "energies"
        and "trajectory" and whatever other managers have been added.

        schedule must be one of "none", "linear","logarithmic","loglin",
        "limlin"

        extra keyword arguments may/must be supplied where relevant, e.g.
        interval=100 for linear scheduling (required).
        base=10 for logarithmic scheduling (optional, default base is 1)
        base=1, maxInterval=16 for loglin (required)
        interval=5, numitems=100 for limlin (required)
        """
        if manager_name not in self.allowedManagers:
            raise ValueError("SetOutputScheduling -- first argument must" \
                  " be one of " + str(self.allowedManagers))

        param_names = {"none":[],
                       "linear":["interval"],
                       "logarithmic":["base"],
                       "loglin":["base", "maxInterval"],
                       "limlin":["interval", "numitems"]}

        if schedule not in param_names.keys():
            raise ValueError("SetOutputScheduling -- second argument " \
                  "must be " + str(param_names.keys()))

        match_pyManager = [x for x in self.pyOutputList
                           if x.name == manager_name]
        if len(match_pyManager) > 0: # python manager
            match_pyManager[0].SetOutputScheduling(schedule, **kwargs)
        else: # C++ manager
            if schedule == "logarithmic" and "base" not in kwargs:
                kwargs["base"] = 1
            self.outputParameters[manager_name] = tuple(
                [schedule] + [kwargs[name] for name in param_names[schedule]])


    def RegisterCallback(self, manager_name, function, **kwargs):
        """
        Register a data analysis function (which takes a Sample as argument)
        to be called with a specified time-step interval. Callback functions
        that are used to generate output should return a string without newline
        characters.
        """
        if manager_name not in self.allowedManagers[2:]:
            raise ValueError("SetOutputParameters -- first argument must" \
                  " be one of" + str(self.allowedManagers[2:]))
        for pyManager in self.pyOutputList:
            if pyManager.name == manager_name:
                pyManager.AddCallback(function, **kwargs)


    def SetRuntimeAction(self, name, method, interval):
        """
        Specify that a user-supplied bound method, taking no arguments,
        to be called during the main loop (before the force calculation)
        every interval time steps
        """
        self.runtime_actions[name] = (
            weakref.ref(method.__self__), weakref.ref(method.__func__))
        # (use the python3 way instead of older im_self and im_func)
        self.runtime_action_intervals[name] = interval

    def RemoveRuntimeAction(self, name):
        """
        Remove an item from the list of runtime-actions
        """
        if name in self.runtime_actions:
            del self.runtime_actions[name]
            del self.runtime_action_intervals[name]
        else:
            print("[Info] Tried to remove non-existent runtime" \
                  " action %s" % str(name))

    def ApplyRuntimeActions(self, time_step_index):
        """
        Called during main loop. Not to be called by the user.
        """
        for key in self.runtime_action_intervals.keys():
            weak_obj, weak_func = self.runtime_actions[key]
            interval = self.runtime_action_intervals[key]
            func = weak_func()
            obj = weak_obj()
            if interval > 0 and time_step_index % interval == 0:
                # func is not bound to obj, so we pass obj as the first
                # (here only) argument (the corresponding "self")
                func(obj)

    def SetMomentumResetInterval(self, mom_reset_interval):
        """
        Set how many time steps should go between resetting of
        center of mass momentum to zero.
        """
        self.runtime_action_intervals["momentum_reset"] = mom_reset_interval

    def ResetMomentum(self):
        """
        Sets the total momentum to zero via a Galilean velocity transformation
        """
        # It is necessary to make this function in order to be correctly
        # include it in the runtime_actions dictionary (since at construction,
        # before the integrator has been set, self.sample.GetIntegrator()
        # returns None
        self.sample.GetIntegrator().SetMomentumToZero()

    def SetSortInterval(self, sort_interval):
        """
        Set how many time steps should go between sorting of
        particle positions (default value 200). Not necessary to be
        called by user if using autotune.
        """
        self.runtime_action_intervals["sort"] = sort_interval

    def InitializeOutput(self, num_iter):
        """
        Initialize output managers at start of Run(),
        including choosing block size if not set by user.
        Not necessary to be called by the user.
        """
        num_active = 0
        min_blockSize = 1
        for manager_name, outputParameters in self.outputParameters.items():
            # C++ managers
            if outputParameters[0] == "none":
                self.sample.SetOutputManagerActive(manager_name, False)
                continue
            num_active += 1
            self.sample.SetOutputManagerActive(manager_name, True)

            schedule = outputParameters[0]
            if schedule == "linear" or schedule == "limlin":
                interval = outputParameters[1]
                user_maxIndex = -1
                if schedule == "limlin":
                    user_maxIndex = outputParameters[2]-1

                self.sample.SetLogLinParameters(manager_name, interval,
                                                interval, user_maxIndex)
                if interval > min_blockSize:
                    min_blockSize = interval
            elif schedule == "logarithmic":
                self.sample.SetLogLinParameters(manager_name,
                                                base=outputParameters[1],
                                                maxInterval=0)
            elif schedule == "loglin":
                self.sample.SetLogLinParameters(
                    manager_name,
                    base=outputParameters[1],
                    maxInterval=outputParameters[2])
            else:
                raise ValueError("Simulation.InitializeOutput: Unknown schedule %s" % schedule)
            
                
        if self.blockSize is None:
            if num_active > 0:
                nominalBlockSize = num_iter//self.nominalNumBlocks//2
                if min_blockSize < nominalBlockSize:
                    min_blockSize = nominalBlockSize
                blockSize = 1
                while blockSize < min_blockSize:
                    blockSize *= 2
            else:
                blockSize = num_iter//self.nominalNumBlocks

            self.blockSize = blockSize
        numBlocks = num_iter//self.blockSize + 1

        if self.verbose:
            print("Output blockSize=%d, numBlocks=%d" % (self.blockSize,
                                                         numBlocks))
            print("Size of the last incomplete block", num_iter % self.blockSize)

        # If blocksize is not set for python output managers it's the same.
        self.sample.SetOutputBlockSize(self.blockSize)
        for pyManager in self.pyOutputList:
            if pyManager.blockSize is None:
                pyManager.SetBlockSize(self.blockSize)
            pyManager.Initialize() # reset counters after any previous run
                
        self.output_is_initialized = True

    def CheckSortingClash(self):
        """ Check whether predefined sorting interval has been set
        when sorting based neighbor-list construction is being used"""
        NB_using_sort = False
        for pot in self.potentialList:
            try:
                if pot.GetNB_Method() == "sort":
                    NB_using_sort = True
            except AttributeError:
                continue
        if NB_using_sort and self.runtime_action_intervals["sort"] > 0:
            print("[Warning] sort_interval for ordinary sorting should be" \
                  " zero when using the sort-based NB method. Setting to" \
                  " zero now.")
            self.SetSortInterval(0)


    def NoteTime(self):
        """ Note elapsed time since since last call to this function
        or since creation of simulation object """

        time_stamp = rumd.Device.GetDevice().Time()
        self.elapsed = time_stamp - self.last_time_stamp
        self.last_time_stamp = time_stamp


    def Run(self, num_iter, initializeOutput=True,
            restartBlock=None, suppressAllOutput=False, force_timing=False):
        """
        Run num_iter time steps, by default initializing/resetting output.
        Specify restartBlock to restart a simulation which has been interrupted.
        Set suppressAllOutput to True to avoid any output (for example for
        equilibration).
        """

        if self.itg is None:
            raise ValueError("Integrator has not been set")
        if len(self.potentialList) == 0:
            raise ValueError("Potential has not been set")

        self.CheckSortingClash()

        initialTimeStep = 0
        if not suppressAllOutput:
            if initializeOutput:
                self.InitializeOutput(num_iter)
            else:
                if not self.output_is_initialized:
                    raise ValueError("Cannot have initializeOutput=False " \
                          "if output has not previously been initialized")

            if restartBlock is not None:
                if not initializeOutput:
                    raise ValueError("When restarting, initializeOutput " \
                          "must be True")
                self.sample.ReadRestartConf(restartBlock)
                initialTimeStep = restartBlock*self.blockSize
        # for timing
        timing = (not suppressAllOutput) or force_timing
        if timing:
            self.NoteTime()

        # main loop
        try:
            iterator = xrange(initialTimeStep, num_iter)
        except NameError:
            iterator = range(initialTimeStep, num_iter)
        for i in iterator:

            self.ApplyRuntimeActions(i)

            # Calculate forces
            need_stress = self.itg.RequiresStress()
            self.sample.CalcF(need_stress)
            
            if not suppressAllOutput:
                # Called here to ensure synchronization of positions and forces
                self.sample.NotifyOutputManagers(i)

                for manager in self.pyOutputList: # python managers for callback
                    if manager.schedule != "none":
                        if manager.nextCalcTimeStep == i:
                            manager.CalcOutput(self.sample)
                        if manager.nextWriteTimeStep == i:
                            manager.WriteOutput(self.sample)
            self.itg.Integrate() # Update positions and velocities
        # end main loop

        if timing:
            self.NoteTime()
            if self.write_timing_info:
                with open(self.timing_filename, "a") as timing_file:
                    nsteps = num_iter-initialTimeStep
                    timing_file.write("%d %.3f %.3f\n" % (
                        nsteps, self.elapsed,
                        self.sample.GetNumberOfParticles()*nsteps/
                        1e6/self.elapsed))

        rumd.Device.GetDevice().CheckErrors()

        # empty last buffers with data
        if not suppressAllOutput:
            for manager in self.pyOutputList:
                if manager.output != []:
                    manager.WriteOutput(self.sample)

        return self.elapsed
