#!/usr/bin/env python
"""Checks that interactions are correctly calculated with more than 5 types"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import unittest
import os
import numpy
import sys
import tempfile
import random

py_rumd_lib = {2:"../../../lib", 3:"../../../lib_py3"}[sys.version_info[0]]
sys.path.insert(1, py_rumd_lib)
sys.path.insert(1, "../../../Python")


import rumd
import rumd.Simulation
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)


def CreateLattice(nSites, latt_a):
    """Create a cubic lattice with given num. particles and lattice spacing """
    num_lattice_side = int(pow(nSites, 1./3)) + 1

    candidate_pos = []
    for idx in range(num_lattice_side):
        for jdx in range(num_lattice_side):
            for kdx in range(num_lattice_side):
                this_pos = [idx, jdx, kdx]
                pos_item = numpy.array(this_pos)*latt_a
                candidate_pos.append(pos_item)


    sorted_pos = sorted(candidate_pos, key=lambda p: numpy.sum(p**2))

    # sort by distance then take the first nSites of them

    pos = numpy.zeros((nSites, 3))
    for pdx in range(nSites):
        pos[pdx, :] = sorted_pos[pdx]

    return pos

def CreateXYZ_String(L, positions, types):
    """Create string for xyz file given box length and positions """
    N = len(positions)
    nTypes = len(types)
    if N != nTypes:
        raise ValueError("There should be as many positions as types")
    
    xyz = "%d\nioformat=2 " % N
    xyz += "sim_box=RectangularSimulationBox,%.4f,%.4f,%.4f " % (L, L, L)
    xyz += "numTypes=%d  " % nTypes
    xyz += "mass=1.000"
    for t in range(nTypes-1):
        xyz += ",1.000"
    xyz += " columns=type,x,y,z\n"

    for t, this_pos in zip(types, positions):
        xyz += "%d %.3f %.3f %.3f\n" % (t,
                                        this_pos[0],
                                        this_pos[1],
                                        this_pos[2])
    return xyz

def vLJ(sigma, epsilon, r):
    """ Evaluate the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(pow(r/sigma, -12.0) - pow(r/sigma, -6.0))

def vLJ_force(sigma, epsilon, r):
    """ Evaluate the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(12.*pow(r/sigma, -12.0) - 6.0*pow(r/sigma, -6.0)) / r

def vLJ_virial(sigma, epsilon, r):
    """Evaluate the virial for the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(4.*pow(r/sigma, -12.0) - 2.*pow(r/sigma, -6.0))

class TestManyTypes(unittest.TestCase):
    """Check that cutoff for unlike particles is correctly handled """
    def setUp(self):

        nTypes = 20
        self.pos = CreateLattice(nTypes, 1.0)
        N = len(self.pos)
        self.types = list(range(nTypes))
        self.cutoff_factor = 2.5


        L = 4 * int(pow(N, 1./3))
        xyz_str = CreateXYZ_String(L, self.pos, self.types)

        self.tdir = tempfile.mkdtemp()
        self.finp = self.tdir + '/test_cutoff_bug.xyz'
        fh = open(self.finp, 'w')
        fh.write(xyz_str)
        fh.close()

        self.sim = rumd.Simulation.Simulation(self.finp, verbose=False)

        self.sigma = {}
        self.epsilon = {}
        self.vCut = {}
        self.fCut = {}
        for idx in range(N):
            for jdx in range(idx, N):
                self.sigma[(idx, jdx)] = random.uniform(0.9, 1.2)
                self.epsilon[(idx, jdx)] = random.uniform(0.5, 2.0)

                self.vCut[(idx, jdx)] = vLJ(self.sigma[(idx, jdx)],
                                            self.epsilon[(idx, jdx)],
                                            self.sigma[(idx, jdx)]* \
                                            self.cutoff_factor)
                self.fCut[(idx, jdx)] = vLJ_force(self.sigma[(idx, jdx)],
                                                  self.epsilon[(idx, jdx)],
                                                  self.sigma[(idx, jdx)] * \
                                                  self.cutoff_factor)


    def SetPotential(self, cutoff_method):

        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=cutoff_method)
        # Set up parameters for Lennard-Jones interactions
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
        potential.SetVerbose(False)
        N = len(self.pos)
        for idx in range(N):
            for jdx in range(idx, N):
                potential.SetParams(idx, jdx, self.sigma[(idx, jdx)],
                                    self.epsilon[(idx, jdx)],
                                    self.cutoff_factor)


        self.sim.SetPotential(potential)



    def test_many_types_pe_shifted_pot(self):
        """Calculate potential energy (shifted pot) by hand, compare to rumd"""

        self.SetPotential(rumd.ShiftedPotential)

        pe = 0.0
        N = len(self.pos)
        for idx in range(N):
            for jdx in range(idx+1, N):
                type_i = self.types[idx]
                type_j = self.types[jdx]

                ordered_types = (type_i, type_j)
                if type_i > type_j:
                    ordered_types = (type_j, type_j)

                sigma = self.sigma[ordered_types]
                epsilon = self.epsilon[ordered_types]
                vc = self.vCut[ordered_types]

                r = numpy.sqrt(numpy.sum((self.pos[idx] - self.pos[jdx])**2))
                if r < sigma*self.cutoff_factor:
                    pe += vLJ(sigma, epsilon, r) - vc

        self.sim.sample.CalcF()
        pe_rumd = self.sim.sample.GetPotentialEnergy()

        #print pe, pe_rumd
        self.assertLess(abs((pe_rumd-pe)/pe), 1.e-6)


    def test_many_types_pe_shifted_for(self):
        """ Calculate potential energy by hand and compare to rumd CalcF"""

        self.SetPotential(rumd.ShiftedForce)
        pe = 0.0
        N = len(self.pos)
        for idx in range(N):
            for jdx in range(idx+1, N):
                type_i = self.types[idx]
                type_j = self.types[jdx]

                ordered_types = (type_i, type_j)
                if type_i > type_j:
                    ordered_types = (type_j, type_j)

                sigma = self.sigma[ordered_types]
                epsilon = self.epsilon[ordered_types]

                vc = self.vCut[ordered_types]
                fc = self.fCut[ordered_types]
                rc = sigma * self.cutoff_factor

                r = numpy.sqrt(numpy.sum((self.pos[idx] - self.pos[jdx])**2))
                if r < rc:
                    pe += vLJ(sigma, epsilon, r) + fc*(r-rc) - vc

        self.sim.sample.CalcF()
        pe_rumd = self.sim.sample.GetPotentialEnergy()

        #print pe, pe_rumd
        self.assertLess(abs((pe_rumd-pe)/pe), 1.e-6)


    def test_many_types_vir_shifted_pot(self):
        """ Calculate virial by hand, compare to rumd CalcF"""

        self.SetPotential(rumd.ShiftedPotential)

        vir = 0.0
        N = len(self.pos)
        for idx in range(N):
            for jdx in range(idx+1, N):
                type_i = self.types[idx]
                type_j = self.types[jdx]

                ordered_types = (type_i, type_j)
                if type_i > type_j:
                    ordered_types = (type_j, type_j)

                sigma = self.sigma[ordered_types]
                epsilon = self.epsilon[ordered_types]

                r = numpy.sqrt(numpy.sum((self.pos[idx] - self.pos[jdx])**2))
                if r < sigma*self.cutoff_factor:
                    vir += vLJ_virial(sigma, epsilon, r)

        self.sim.sample.CalcF()
        vir_rumd = self.sim.sample.GetVirial()

        #print vir, vir_rumd
        self.assertLess(abs((vir_rumd-vir)/vir), 1.e-6)

    def tearDown(self):
        import shutil
        if os.path.exists(self.finp):
            os.remove(self.finp)
        if os.path.exists(self.tdir):
            shutil.rmtree(self.tdir)


if __name__ == '__main__':
    unittest.main()
